from os import  environ, makedirs
from os.path import isfile, join, isdir
from typing import List
from ipaddress import ip_address
from toml import load as toml_load
from json import load, dump
from pprint import pprint


MULTI_FUNCTIONS = []
EXTRA_ANNOTATORS = ['risotto.rougail']
ROUGAIL_NAMESPACE = 'general'
ROUGAIL_NAMESPACE_DESCRIPTION = 'Général'
HERE = environ['PWD']
IP_DIR = join(HERE, 'ip')


config_file = environ.get('CONFIG_FILE', 'risotto.conf')
if isfile(config_file):
    with open(config_file, 'r') as fh:
        RISOTTO_CONFIG = toml_load(fh)
else:
    RISOTTO_CONFIG = {}


def _(s):
    return s


def multi_function(function):
    global MULTI_FUNCTIONS
    name = function.__name__
    if name not in MULTI_FUNCTIONS:
        MULTI_FUNCTIONS.append(name)
    return function


async def value_pprint(dico, config):
    pprint_dict = {}
    for path, value in dico.items():
        if await config.option(path).option.type() == 'password' and value:
            value = 'X' * len(value)
        pprint_dict[path] = value
    pprint(pprint_dict)


def load_zones(zones, hosts):
    if not isdir(IP_DIR):
        makedirs(IP_DIR)
    json_file = join(IP_DIR, 'zones.json')
    if isfile(json_file):
        with open(json_file, 'r') as fh:
            zones_ip = load(fh)
    else:
        zones_ip = {}
    for host_name, hosts in hosts.items():
        for server_name, server in hosts['servers'].items():
            server_zones = server['informations']['zones_name']
            for idx, zone_name in enumerate(server_zones):
                zone = zones[zone_name]
                zone.setdefault('hosts', {})
                # FIXME make a cache, machine should not change IP
                if zone_name not in zones_ip:
                    zones_ip[zone_name] = {}
                if server_name in zones_ip[zone_name]:
                    server_index = zones_ip[zone_name][server_name]
                elif not zones_ip[zone_name]:
                    server_index = 0
                else:
                    # it's the last ip + 1
                    server_index = zones_ip[zone_name][list(zones_ip[zone_name].keys())[-1]] + 1
                ip = str(ip_address(zone['start_ip']) + server_index)
                zone['hosts'][server_name] = ip
                zones_ip[zone_name][server_name] = server_index
    with open(json_file, 'w') as fh:
        dump(zones_ip, fh)
